-------------------------------------------------------------------------------
-- Common utility functions and components
--
-- Functions:
--   - max  : integer maximum
--   - min  : integer minimum
--   - log2 : floor of base-2 logarithm
--
-- Components:
--   - mux : multiplexer
-------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;

package pkg_common is

  -----------------------------------------------------------------------------
  -- Integer maximum function
  -----------------------------------------------------------------------------

  function max ( x, y : integer ) return integer;

  -----------------------------------------------------------------------------
  -- Integer minimum function
  -----------------------------------------------------------------------------

  function min ( x, y : integer ) return integer;

  -----------------------------------------------------------------------------
  -- Floor of base-2 logarithm
  -----------------------------------------------------------------------------

  function log2 ( x : natural ) return natural;

  -----------------------------------------------------------------------------
  -- Multiplexer
  -----------------------------------------------------------------------------

  component mux is
    generic ( n : positive );
    port ( i : in  std_logic_vector(2**n-1 downto 0);
           a : in  std_logic_vector(n-1 downto 0);
           o : out std_logic );
  end component;

  -----------------------------------------------------------------------------
  -- Left shift operator
  -----------------------------------------------------------------------------

  component left_shift is
    generic ( w : positive;
              n : positive );
    port ( i : in  std_logic_vector(w-1 downto 0);
           s : in  std_logic_vector(n-1 downto 0);
           o : out std_logic_vector(w+2**n-2 downto 0) );
  end component;

  -----------------------------------------------------------------------------
  -- Right shift operator
  -----------------------------------------------------------------------------

  component right_shift is
    generic ( w : positive;
              n : positive );
    port ( i : in  std_logic_vector(w-1 downto 0);
           s : in  std_logic_vector(n-1 downto 0);
           o : out std_logic_vector(w+2**n-2 downto 0) );
  end component;

  -----------------------------------------------------------------------------
  -- Leading-zero counter and normalization
  -----------------------------------------------------------------------------

  component lzc_norm is
    generic ( w : positive;
              n : positive );
    port ( i : in  std_logic_vector(w-1 downto 0);
           z : out std_logic_vector(n-1 downto 0);
           o : out std_logic_vector(w-1 downto 0) );
  end component;

end package;

-------------------------------------------------------------------------------

package body pkg_common is

  -----------------------------------------------------------------------------
  -- Integer maximum function
  --
  -- Inputs:
  --   - x [integer] : number
  --   - y [integer] : number
  --
  -- Return:
  --   - [integer] : the largest integer between x and y
  -----------------------------------------------------------------------------
  
  function max ( x, y : integer ) return integer is
  begin
    if x > y then
      return x;
    else
      return y;
    end if;
  end function;

  -----------------------------------------------------------------------------
  -- Integer minimum function
  --
  -- Inputs:
  --   - x [integer] : number
  --   - y [integer] : number
  --
  -- Return:
  --   - [integer] : the smallest integer between x and y
  -----------------------------------------------------------------------------
  
  function min ( x, y : integer ) return integer is
  begin
    if x < y then
      return x;
    else
      return y;
    end if;
  end function;

  -----------------------------------------------------------------------------
  -- Floor of base-2 logarithm
  --
  -- Inputs:
  --   - x [natural] : number
  --
  -- Return:
  --   - [natural] : floor of the base-2 logarithm of x
  -----------------------------------------------------------------------------
  
  function log2 ( x : natural ) return natural is
    variable n : natural := 0;
  begin
    while 2**(n+1) <= x loop
      n := n+1;
    end loop;
    return n;
  end function;

end package body;
