-------------------------------------------------------------------------------
-- Floating-point sine/cosine (dual operator)
--
-- Generics:
--   - wE [positive] : exponent bit-width
--   - wF [positive] : fraction bit-width
--
-- Ports:
--   - fpX [in]  : operand X in FP format
--   - fpS [out] : sine of X in FP format
--   - fpC [out] : cosine of X in FP format
-------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
library fplib;
use fplib.pkg_common.all;
use fplib.pkg_fp_common.all;
use fplib.pkg_fp_sin_cos.all;

entity fp_sin_cos_fxp is
  generic ( wE : positive := 6;
            wF : positive := 13 );
  port ( fpX : in  std_logic_vector(wE+wF+2 downto 0);
         fpS : out std_logic_vector(wE+wF+2 downto 0);
         fpC : out std_logic_vector(wE+wF+2 downto 0) );
end entity;

architecture arch of fp_sin_cos_fxp is
  signal eX : std_logic_vector(wE downto 0);
  signal xX : std_logic_vector(1 downto 0);

  constant gK  : positive := fp_sin_cos_4oPi_g(wE, wF);
  constant nK0 : std_logic_vector(3+2*wF+2**(wE-1)-1+gK+6 downto 0) :=
                                                       fp_sin_cos_4oPi(wE, wF);
  signal   nK1 : std_logic_vector(3+2*wF+2**wE-2+3 downto 0);
  signal   nK2 : std_logic_vector(3+2*wF+2**(wE-1)-1+3 downto 0);
  signal   nK  : std_logic_vector(3+2*wF+3 downto 0);

  signal nY0 : std_logic_vector(3*wF+7 downto 0);
  signal nY1 : std_logic_vector(2+wF+3 downto 0);
  signal nY  : std_logic_vector(2+wF+2 downto 0);

  signal far_nZ  : std_logic_vector(wF+1 downto 0);

  signal cls_nZ0 : std_logic_vector(wF+2+2**log2(wF+3)-1 downto 0);
  signal cls_nZ1 : std_logic_vector(wF+2 downto 0);
  signal cls_nZ  : std_logic_vector(wF+1 downto 0);
  signal cls_eZ0 : std_logic_vector(wE-1 downto 0);

  signal nZ : std_logic_vector(wF+1 downto 0);

  signal nS0 : std_logic_vector(wF+2 downto 0);
  signal nS1 : std_logic_vector(wF downto 0);
  signal mS1 : std_logic_vector(wF downto 0);
  signal eS1 : std_logic_vector(log2(wF+1)-1 downto 0);
  signal xS  : std_logic_vector(1 downto 0);
  signal eS  : std_logic_vector(wE-1 downto 0);
  signal fS  : std_logic_vector(wF-1 downto 0);

  signal nC0 : std_logic_vector(wF+2 downto 0);
  signal mC1 : std_logic_vector(wF+1 downto 0);
  signal eC  : std_logic_vector(wE-1 downto 0);
  signal fC  : std_logic_vector(wF-1 downto 0);

begin --------------------------------------------------------- Range reduction

  eX <= ("0" & fpX(wE+wF-1 downto wF)) - ("00" & (wE-2 downto 1 => '1') & "0");

  shift_nk0 : left_shift
    generic map ( w => 3+2*wF+2**(wE-1)-1+4,
                  n => wE-1 )
    port map ( i => nK0(nK0'high downto 1+gK+2),
               s => eX(wE-2 downto 0),
               o => nK1 );

  nK2 <= nK1(nK1'high downto 2**(wE-1)-1);

  nK <= nK2(3+2*wF+3 downto 0)                   when eX(wE-1) = '0' else
        nK0(3+2*wF+gK+5 downto gK+2)             when eX(wE) = '0'   else
        nK0(nK0'high downto nK0'high-(3+2*wF+3));

  nY0 <= ("1" & fpX(wF-1 downto 0)) * nK;
  nY1 <= nY0(3+2*wF+3 downto wF+1);
  nY  <= nY1(2+wF+3 downto 1) + ((2+wF+2 downto 1 => '0') & nY1(0));

  --------------------------------------------------- Range reduction, far path
  
  far_nZ <= nY(wF+1 downto 0)                          when nY(wF+2) = '0' else
            (wF+1 downto 0 => '0') - nY(wF+1 downto 0);
  -- problem when nY = 1.0

  ------------------------------------------------- Range reduction, close path

  cls_eZ0 <= (wE-1 downto 0 => '0') - eX(wE-1 downto 0);
  
  shift_cls_nz : right_shift
    generic map ( w => wF+3,
                  n => log2(wF+3) ) -- problem when eZ0 is too large
    port map ( i => nY1(wF+3 downto 1),
               s => cls_eZ0(log2(wF+3)-1 downto 0),
               o => cls_nZ0 );

  cls_nZ1 <= cls_nZ0(wF+1+2**log2(wF+3)-1 downto -1+2**log2(wF+3)-1);
  cls_nZ  <= cls_nZ1(wF+2 downto 1) + ((wF+1 downto 1 => '0') & cls_nZ1(0));

  ----------------------------------------------------- Range reduction, select

  nZ <= far_nZ when eX(wE-1) = '0' else
        cls_nZ;

  ------------------------------------------------------------------------ Sine

  sin_nz : fp_sin_cos_sin
    generic map ( wF => wF )
    port map ( x => nZ,
               r => nS0 );

  nS1 <= nS0(wF+1 downto 1) + ((wF downto 1 => '0') & nS0(0));

  lzc_norm_ns : lzc_norm
    generic map ( w => wF+1,
                  n => log2(wF+1) )
    port map ( i => nS1,
               z => eS1,
               o => mS1 );

  xS <= "0" & mS1(wF);
  eS <= ("0" & (wE-2 downto 1 => '1') & "0") - ((wE-1 downto log2(wF+1) => '0') & eS1(log2(wF+1)-1 downto 0));
  fS <= mS1(wF-1 downto 0);

  ---------------------------------------------------------------------- Cosine

  cos_nz : fp_sin_cos_cos
    generic map ( wF => wF )
    port map ( x => nZ,
               r => nC0 );

  mC1 <= nC0(wF+2 downto 1) + ((wF+1 downto 1 => '0') & nC0(0));

  eC <= "0" & (wE-2 downto 1 => '1') & mC1(wF+1);
  fC <= (wF-1 downto 0 => '0') when mC1(wF+1) = '1' else
        mC1(wF-1 downto 0);

  -----------------------------------------------------------------------------

  xX <= fpX(wE+wF+2 downto wE+wF+1);

  fpS(wE+wF+2 downto wE+wF+1) <= "00" when xX = "00" else
                                 "11" when xX(1) = '1' else
                                 xS   when nY(wF+2) = nY(1+wF+2) or eX(wE) = '1' else
                                 "01";

  fpS(wE+wF) <= fpX(wE+wF) when eX(wE) = '1' else
                fpX(wE+wF) xor nY(2+wF+2);

  fpS(wE+wF-1 downto 0) <= eS & fS when nY(wF+2) = nY(1+wF+2) or eX(wE) = '1' else
                           eC & fC;

  fpC(wE+wF+2 downto wE+wF+1) <= xX(1) & "1";

  fpC(wE+wF) <= fpX(wE+wF) when eX(wE) = '1' else
                fpX(wE+wF) xor nY(2+wF+2) xor nY(1+wF+2);

  fpC(wE+wF-1 downto 0) <= eC & fC when nY(wF+2) = nY(1+wF+2) or eX(wE) = '1' else
                           eS & fS;

end architecture; -------------------------------------------------------------
